function [A,labels] = bigblock(ns,ps,p,varargin)
% BIGBLOCK Create a big block model.
%
% A = bigblock(ns,ps,p) creates a graph block model
%       where block 1 has ns(1) nodes with internal connection prob ps(1)
%       where block 2 has ns(2) nodes with internal connection prob ps(2)
%       ...
%       where block k has ns(end) nodes and internal connection prob ps(end)
% and where the global connection prob is p.
%
% [A,labels] = bigblock(...) also returns the id of each block.

parser = inputParser;
parser.addParamValue('weighted',false,@islogical);
parser.addParamValue('badsparse',false,@islogical);
parser.parse(varargin{:});
opts = parser.Results;

nblocks = numel(ns);
nverts = sum(ns);
blocks = cell(nblocks,1);
labels = zeros(nverts,1);
blockstart = 1;

for i=1:nblocks
    B = sprand(ns(i),ns(i),ps(i));
    B = triu(B,1);
    B = B + B';
    if opts.badsparse
        cmax = max(B); % max in each column
        wmin = cmax.*rand(ns(i)); % find the min element we'll accept
        [ei,ej,ev] = find(B); % get the sparity structure
        f = ev > wmin(ej); % create the filter for the non-zeros
        B = sparse(ei,ej,ev,ns(i),ns(i));
        B = max(B,B');
    end
    if ~opts.weighted
        B = spones(B);
    else
        B = min(2*B,spones(B)); % increase weights
    end
    blocks{i} = B;
    labels(blockstart:blockstart+ns(i)-1) = i;
    blockstart = blockstart + ns(i);
end
A = blkdiag(blocks{:});
R = sprand(nverts,nverts,p);
R = triu(R,1);
R = R + R';
if ~opts.weighted
    R = spones(R);
end
A = max(A,R);

% trim off empty rows/cols
d = sum(A,2);
valid = d > 0;
A = A(valid,valid);
labels = labels(valid);