%% Use the USPS data for semi-supervised learning
% This experiment replicates the one done in Zhou 2003 on semi-supervised
% learning on a graph.

%% Load the data
load data/usps.mat
solve_cvx = 0; % = 0 means to use gurobi, = 1 means use (SLOW) cvx

%% digit subset
digits = [1,2,3,4]+1; % handle the offset that 1 = 0
patterns = [];
labels = [];
for d = digits
    filt = train_labels(d,:)==1;
    labels = [labels d*ones(1,sum(filt))];
    patterns = [patterns train_patterns(:,filt)];
    filt = test_labels(d,:)==1;
    labels = [labels d*ones(1,sum(filt))];
    patterns = [patterns test_patterns(:,filt)];
end
labels = labels-1; % handle the offset 

%% construct the graph
D = pdist2(patterns',patterns','euclidean');
%%
rad = 1.25;
K = exp(-D.^2/(2*rad^2));
Gf = K - diag(diag(K));
G = Gf;
G(G<exp(-10)) = 0; % determined this threshold by hand
G = sparse(G);
n = size(K,1);
d = sum(G,2);
% Patch up the empty rows
G(:,d==0) = Gf(:,d==0);
G(d==0,:) = Gf(d==0,:);
d = sum(G,2);
W = diag(1./sqrt(d))*G*diag(1./sqrt(d));
P = G*diag(1./d);
%%
beta = 0.99;
alpha = (1-beta)/beta;
% R = chol(K) => R'*R = K, so Kx = Y => R'*Rx = Y => Rx = (R'\Y)
K = full(eye(n) - beta*W); R = chol(K); iK1 = @(Y) R \ (R' \ Y); 
K = full(diag(d) - beta*G); R = chol(K); iK2 = @(Y) R \ (R' \ Y); 
%%
ntrials = 5;
nlabeled = [4 10:10:50];
nl = numel(nlabeled);
classes = unique(labels);
nclasses = numel(classes);

assert(numel(labels) == n);
errrate = zeros(nl, ntrials, 3);
for j=1:nl
    nlabels = nlabeled(j)
    for i=1:ntrials
        %%
        % pick the training set, it needs to have one element from each class.
        p = randperm(n);
        train_set = zeros(n,1);
        pl = labels(p);
        [cp,ia] = unique(pl,'first'); %pl(ia) = c
        train_set(p(ia)) = 1;
        test_set = ~train_set;
        test_labels = labels(test_set)';


        p(ia) = [];
        train_set(p(1:(nlabels - nclasses))) = 1;

        Y = full(sparse(find(train_set),labels(logical(train_set)),1,n,nclasses));
        Yt = diag(1./sqrt(d))*Y;       

        Pred = iK1(Y);
        [~,Pi] = max(Pred,[],2); % find the largest entry in each row
        pred_labels = Pi(test_set);
        errs = numel(test_set) - sum(pred_labels == test_labels);
        errrate(j,i,1) = errs/numel(test_set);
        
        Pred = iK2(Y);
        [~,Pi] = max(Pred,[],2); % find the largest entry in each row
        pred_labels = Pi(test_set);
        errs = numel(test_set) - sum(pred_labels == test_labels);
        errrate(j,i,2) = errs/numel(test_set);
        
        % Predict using cuts
        Pred = zeros(size(Y));
        for c=1:nclasses
            % Run 1-norm SSL for our example
            % v is the "seed vector"
            s = Yt(:,c);
            sbar = d - s;
            assert(all(sbar >= -eps(1)));
            assert(all(s >= 0));
            
            % Okay, we can't do the full 1-norm problem
            if solve_cvx
                AS = [0              alpha*(s)'   0;
                      alpha*(s)  G     alpha*(sbar);
                      0              alpha*(sbar)' 0];

                clear M;
                [ei ej ev] = find(triu(AS,0));
                C = diag(sparse(ev));
                nedge = numel(ei);
                Bu = sparse([(1:nedge)'; (1:nedge)'], [ei ej], ...
                    [ones(nedge,1); -1*ones(nedge,1)], ...
                    nedge, size(AS,1));


                cvx_begin
                    variable xt(n+2);
                    minimize norm(C*Bu*xt,1)
                    subject to
                        xt(1) == 1;
                        xt(n+2) == 0;
                cvx_end
            else 
                fprintf('Solving with Gurobi i=%i\n', i);
                xt = prcut_gurobi(G,alpha,s,d);
            end
            Pred(:,c) = sqrt(d).*xt; % doesn't actually change results, but "right"
        end
        [~,Pi] = max(Pred,[],2); % find the largest entry in each row
        pred_labels = Pi(test_set);
        errs = numel(test_set) - sum(pred_labels == test_labels);
        errrate(j,i,3) = errs/numel(test_set);
    end
end

%%
plot(squeeze(mean(errrate,2)),'o-');
legend('Kernel 1', 'Kernel 2', 'Cuts');