function Clow = fisher_z_conf_low(C,N,level)
% FISHER_Z_CONF_LOW Use Fisher's Z transform to get the lower confidence bound
%
% Clow = pearson_confidence(C,N) takes in correlation coefficients and the
% number of associated measurements and returns Clow, the lower-bound
% of a 95% confidence interval on the true correlation. This also works
% with any [-1,1] distributed variable. It's an approximate confidence
% interval.
%
% Clow = pearson_confidence(C,N,level) gives the desired two-sided
% confidence level.
%
% The method is based on Fisher's Z transform, i.e. Z = arctanh(C), use a
% normal confidence interval in Z, and transform back with tanh. The
% standard deviation is sqrt(N-3).

if ~exist('level','var') || isempty(level), level = 0.95; end

normoffset = norminv((1-level)/2);

% compute column-wise, because otherwise it takes too much time!
Clow = zeros(size(C));
for i=1:size(C,2)
    ci = C(:,i);
    ni = N(:,i);
    z = atanh(ci);
    zlow = z + normoffset./sqrt(ni-3).*(ni > 3);
    zlow(ni <= 3) = 0;
    Clow(:,i) = tanh(zlow);
end


% % compute element-wise because otherwise this takes too much memory.
% L = zeros(size(C));
% for i=1:numel(C)
%     if N(i) < 4
%         L(i) = 1;
%     else
%         z = atanh(C(i));
%         conf = -z.*sqrt(N(i)-3);
%         l = min((1-normcdf(conf))*2, 2*normcdf(conf));
%         L(i) = l;
%     end
% end
% Z = atanh(C);
% we need to find the confidence level that excludes 0 
% Z + X./sqrt(N-3) = 
% Conf = -Z .* sqrt(N-3);
% L = min((1-normcdf(Conf))*2, 2*normcdf(Conf));
