function [y,kappa] = prl1_sweep(prl1_solver, maxnodes, minnodes)
% PRL1_SWEEP Perform a sweep over regularization parameters to find
% something that gives a maximum 
%
% [z,kappa] = prl1_sweep(prl1_solver, maxnodes, minnodes)
%
%   prl1_solver is a pagerank l1 solver parameterized by kappa, see the
%   example.
%
% finds the largest value of kappa such that the solution z has fewer
% than maxnodes*n non-zero entries. The algorithm is a bisection search
% that terminates within a factor of two of maxnodes, e.g. you'll get a
% return with nnz(y) in [minnodes,maxnodes]*n where n is the size of the
% graph.
%
% Example:
%   A = load_graph('four-clusters');
%   n = size(A,1);
%   s = zeros(n,1); s(1) = 1; 
%   d = sum(A,2);
%   prl1 = @(kappa) prl1_fast(A,0.9,s,d,kappa);
%   [y,kappa] = prl1_sweep(prl1, 0.5) % find a vector with at most half the
%   nodes set.

if nargin < 2
    maxnodes = 1/2;
end
if nargin < 3
    minnodes = maxnodes/2;
end
assert(minnodes < maxnodes);

kappa = 1;
% is kappa too small?
y = prl1_solver(kappa);
n = numel(y);
nzk = nnz(y)/n;
if nzk > minnodes
    for i=1:20
        kappa = kappa*2;
        y = prl1_solver(kappa);
        nzk = nnz(y)/n;
        if nzk < maxnodes
            break;
        end
    end
end
if nzk > maxnodes
    warning('prl1_sweep:noStart', ...
        'after increasing kappa to %e, there were still too many non-zeros %f (%i/%i)', ...
        kappa, nzk, nnz(y), n);
    return
end

assert(nzk < maxnodes);

% in this case, we can return because we satisfy the stopping
if nzk > minnodes
    return
end

% next, we need to decrease kappa until we get too many.
kappa_search = [];
for i=1:12
    kappa_test = kappa/10;
    y = prl1_solver(kappa_test);
    nzk = nnz(y)/n;
    if nzk < minnodes
        kappa = kappa_test; % continue searching
    elseif nzk > maxnodes
        % we have our search bracket
        kappa_search = [kappa_test, kappa];
        break
    else
        assert(nzk <= maxnodes)
        assert(nzk >= minnodes)
        kappa = kappa_test;
        return;
    end
end

if isempty(kappa_search)
    warning('prl1_sweep:noKappa', ...
        'after decreasing kappa to %e, there were still too few non-zeros %f (%i/%i)', ...
        kappa, nzk, nnz(y), n);
    return
end

assert(~isempty(kappa_search))

% Now do bisection search on kappa_search in order to satisfy our 
% criteria minnodes <= nzk <= maxnodes

while diff(kappa_search) > sqrt(eps)
    kappa = mean(kappa_search);
    y = prl1_solver(kappa);
    nzk = nnz(y)/n;
    
    if nzk > maxnodes
        kappa_search(1) = kappa;
    elseif nzk < minnodes
        kappa_search(2) = kappa;
    else
        assert(nzk <= maxnodes)
        assert(nzk >= minnodes)
        return;
    end
end

kappa = kappa_search(2);
y = prl1_solver(kappa);



